/***************************************************************************
 *
 * Copyright (C) 2001 International Business Machines
 * All rights reserved.
 *
 * This file is part of the GPFS mmfslinux kernel module.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer. 
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 *  3. The name of the author may not be used to endorse or promote products 
 *     derived from this software without specific prior written
 *     permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *************************************************************************** */
/* @(#)69       1.61  src/avs/fs/mmfs/ts/DirIds.h, mmfs, avs_rgpfs24, rgpfs24s001a 4/3/06 18:25:52 */
#ifndef _h_DirIds
#define _h_DirIds

/*
 * Define unique numeric IDs for each source directory in Shark that contains
 * code.  These directory IDs are used for several purposes:
 *   1. Generating unique labels ("hookids") for kernel trace points
 *   2. Identifying message set numbers for the message catalog code
 * The symbols defined in this file are formed by appending the relative
 * directory name to the string "DIR_ID_", then translating slashes to
 * underscores.
 *   The values of the directory IDs must be in the range 1...63:
 * zero is not allowed because 0 is not valid as a message set number;
 * values larger than 63 cannot be used because only 6 bits of the ID are
 * used in generating kernel trace hookids.
 *
 * ==========================================================================
 *
 * DO NOT renumber these IDs, or reuse an ID for a directory that
 * has been deleted, as numbers derived from these IDs appear in the
 * message catalog and trace format file.
 *
 * The message files are listed here with their message number ranges.
 * The range of numbers of MMFS messages 3-digit codes
 * reserved for each existing message catalog file (.msg file)
 *
 * The messages displayed have the format "GPFS: 6027-xxx <message text>"
 * where xxx is the unique 3-digit or 4-digit code for the particular error.
 *
 *
 * SUBDIRECTORY            CATALOG FILE NAME                RANGE OF CODES
 * ------------            -----------------                --------------
 *                                                    unused    000-299
 *  mmfsd                     mmfsd.msg                         300-309
 *  tasking (1)                task.msg                         310-339
 *  classes/tscomm           tscomm.msg                         340-359
 *  disk (1)                   disk.msg                         360-379
 *  classes/basic              base.msg                         380-399
 *  stripe                   stripe.msg                         400-499
 *  vfs                        svfs.msg                         500-529
 *  fs                           fs.msg                         530-599
 *  cfgmgr (1)                 cfgm.msg                         600-649
 *  logger                   logger.msg                         650-659
 *  cfgmgr (2)                 cfgm.msg                         660-799
 *  phoenix (1)             phoenix.msg                         800-849
 *  util                       util.msg                         850-999
 *  admin                     admin.msg                        1000-1399
 *  log                         log.msg                        1400-1449
 *  defrag                   defrag.msg                        1450-1499
 *  disk (2)                   disk.msg                        1500-1529
 *  tasking (2)                task.msg                        1530-1549
 *  admin (2)                 admin.msg                        1550-1699
 *  classes/tscomm (2)       tscomm.msg                        1700-1799
 *  nsd                         nsd.msg                        1800-1849
 *  phoenix (2)             phoenix.msg                        1850-1899
 *  admin (3)                 admin.msg                        1900-2049
 *  bufmgr                   bufmgr.msg                        2049-2099
 *  admin (4)                 admin.msg                        2100-2499
 *  sanergy                 sanergy.msg                        2500-2549
 *  bufmgr/linux            bufmgr_linux.msg                   2550-2575
 *  tasking/linux           tasking_linux.msg                  2576-2599
 *  fs (2)                       fs.msg                        2600-2699
 *  cfgmgr (3)                 cfgm.msg                        2700-2799
 *  pc                           pc.msg                        2800-2899
 *
 * ==========================================================================
 *
 * $Id: DirIds.h,v 1.85 2006/04/03 19:25:58 makaplan Exp $
 *
 * $Log: DirIds.h,v $
 * Revision 1.85  2006/04/03 19:25:58  makaplan
 * Claim message numbers 2800-2899 for policy, ts/pc/pc.msg
 *
 * Revision 1.84  2006/03/10 06:29:58  schmuck
 * Prune check-in history.  No code change.
 *
 * Revision 1.83  2006/03/02 03:25:16  hempel
 * message numbering:  add new message ranges for the fs and cfgmgr directories
 *
 * Revision 1.82  2005/11/15 16:26:42  gjertsen
 * Provide a helpful message when daemon encounters an out of memory
 * situation along with more graceful shutdown (defect 518290).
 *
 * Revision 1.81  2005/08/12 19:39:41  gjertsen
 * Fix for 515602. Provide useful error message and shutdown if allocation
 * fails. Create msg catalog file for bufmgr/linux but reuse MSG_E_NOMEM.
 *
 * Revision 1.80  2005/05/03 16:48:44  makaplan
 * POLICY - integration tweaks
 *
 * Revision 1.79  2005/01/15 00:58:41  mcnabb
 * Remove reference to smit messages
 *
 * [ 74 log messages deleted ]
 *
 * Revision 1.1  1993/07/16  18:09:11  wyllie
 * Add file with unique IDs for each Shark source code directory.
 *
 */

/*
 * IMPORTANT (see also comments at the top of this file):
 *  - names must be of the form DIR_ID_XXX, where XXX = relative dir name
 *  - values must be in the range 1 ... 63
 */
#define DIR_ID_CFGMGR                   1
#define DIR_ID_CLASSES_BASIC            2
#define DIR_ID_CLASSES_MALLOC           3
#define DIR_ID_LOCKMGR                  4
#define DIR_ID_CLASSES_TSCOMM           5
#define DIR_ID_DEBUG                    6
#define DIR_ID_DISK                     7
#define DIR_ID_FS                       8
/*#define DIR_ID_FS_OVERFLOW do not reuse 9*/
/* unused                               10 */
#define DIR_ID_INSTALL                  11
#define DIR_ID_LOGGER                   12
#define DIR_ID_MMFSD                    13
#define DIR_ID_ADMIN                    14
#define DIR_ID_PAGEMGR                  15
/*#define DIR_ID_PIPELINE                 16 */
/*#define DIR_ID_RPC                      17 */
#define DIR_ID_TASKING                  18
#define DIR_ID_VFS                      19
#define DIR_ID_KERNEXT                  20
/*#define DIR_ID_TESTAPPS_MPEG            21 */
/*#define DIR_ID_TESTAPPS_SHPLAYER        22 */
/*#define DIR_ID_TESTAPPS_TEST            23 */
/*#define DIR_ID_TESTAPPS_TESTPP          24 */
#define DIR_ID_LOG                      25
#define DIR_ID_CLASSES_PERFMON          26
#define DIR_ID_UTIL                     27
#define DIR_ID_STRIPE                   28
/*#define DIR_ID_CATMNT                         29 */
#define DIR_ID_PHOENIX                  30
#define DIR_ID_UTIL_SCRIPT              31
#define DIR_ID_ADMIN_SCRIPT             32
#define DIR_ID_TM                       33
#define DIR_ID_DM                       34
#define DIR_ID_DM_KX                    35
#define DIR_ID_DEFRAG                   36
#define DIR_ID_CLASSES_MBPERFMON        37
#define DIR_ID_PFSCK                    38
#define DIR_ID_DIRLINUX                 39
#define DIR_ID_KERNEXT_GPL_LINUX        40
#define DIR_ID_KERNEXT_AIX              41
#define DIR_ID_FS_AIX                   42
#define DIR_ID_FS_LINUX                 43
#define DIR_ID_KERNEXT_GPL_LINUX_I386   44
#define DIR_ID_KERNEXT_GPL_LINUX_IA64   45
#define DIR_ID_NSD                      46
#define DIR_ID_LOGGER_AIX               47
#define DIR_ID_LOGGER_LINUX             48
#define DIR_ID_KERNEXT_LINUX            49
#define DIR_ID_TASKING_AIX              50
#define DIR_ID_TASKING_LINUX            51
#define DIR_ID_PC                       52
#define DIR_ID_SANERGY                  53
#define DIR_ID_SANERGY_KX               54
#define DIR_ID_BUFMGR                   55
#define DIR_ID_BUFMGR_LINUX             56
#define DIR_ID_BUFMGR_AIX               57
#define DIR_ID_DM_KX_LINUX              58
#define DIR_ID_DM_KX_AIX                59
#define DIR_ID_FENCING                  60
#define DIR_ID_KERNEL                   61



/* Add new directory IDs above here */

/* 91 is being used for SMIT help */

/* overrides for default hookid (307) in certain directories */
#define SHARKHOOKID_TM      0x30600000
#define SHARKHOOKID_TASKING 0x30800000
#define SHARKHOOKID_KERNEXT 0x30900000

#endif  /* _h_DirIds */
